import math
import os
import string

import nltk
import sys

FILE_MATCHES = 1
SENTENCE_MATCHES = 1


def main():

    # Check command-line arguments
    if len(sys.argv) != 2:
        sys.exit("Usage: python questions.py corpus")

    # Calculate IDF values across files
    files = load_files(sys.argv[1])
    file_words = {
        filename: tokenize(files[filename])
        for filename in files
    }
    file_idfs = compute_idfs(file_words)

    # Prompt user for query
    query = set(tokenize(input("Query: ")))

    # Determine top file matches according to TF-IDF
    filenames = top_files(query, file_words, file_idfs, n=FILE_MATCHES)

    # Extract sentences from top files
    sentences = dict()
    for filename in filenames:
        for passage in files[filename].split("\n"):
            for sentence in nltk.sent_tokenize(passage):
                tokens = tokenize(sentence)
                if tokens:
                    sentences[sentence] = tokens

    # Compute IDF values across sentences
    idfs = compute_idfs(sentences)

    # Determine top sentence matches
    matches = top_sentences(query, sentences, idfs, n=SENTENCE_MATCHES)
    for match in matches:
        print(match)


def load_files(directory):
    """
    Given a directory name, return a dictionary mapping the filename of each
    `.txt` file inside that directory to the file's contents as a string.
    """

    # initialise empty dict
    files = {}

    # iterate through files in directory
    for file in os.listdir(directory):
        # if file is of type .txt then read contents in utf8 encoding
        if os.path.splitext(file)[1] == '.txt':
            filepath = os.path.join(directory, file)
            with open(filepath, encoding="utf8") as f:
                content = f.read()

            # add filename and content to dict
            files[file] = str(content)

    return files


def tokenize(document):
    """
    Given a document (represented as a string), return a list of all of the
    words in that document, in order.

    Process document by coverting all words to lowercase, and removing any
    punctuation or English stopwords.
    """

    # import stopwords for English language
    stp = nltk.corpus.stopwords.words("english")
    # import ascii punctuation
    pnct = string.punctuation

    # generate tokens
    tokens = nltk.word_tokenize(document)

    # initialise empty list for final tokens
    final = []

    # iterate through tokens, if not in stopwords or punctuation append to final list
    for token in tokens:
        if token not in stp and token not in pnct:
            final.append(token.lower())

    return final


def compute_idfs(documents):
    """
    Given a dictionary of `documents` that maps names of documents to a list
    of words, return a dictionary that maps words to their IDF values.

    Any word that appears in at least one of the documents should be in the
    resulting dictionary.
    """

    # initialise idf dict
    idf = {}

    # all words
    words = []

    # total docs
    n = len(documents)

    # iterate over documents in dictionary to get all words
    for doc in documents:
        for word in documents.get(doc):
            if word not in words:
                words.append(word)

    # calculate count
    for word in words:
        counter = 0
        for doc in documents:
            if word in documents.get(doc):
                counter += 1
        # calculate IDF: ln ( N / nr of documents the word is in)
        idf[word] = (math.log(n/counter))

    return idf


def top_files(query, files, idfs, n):
    """
    Given a `query` (a set of words), `files` (a dictionary mapping names of
    files to a list of their words), and `idfs` (a dictionary mapping words
    to their IDF values), return a list of the filenames of the the `n` top
    files that match the query, ranked according to tf-idf.
    """

    # initialise list
    filenames = []

    # dict to keep track of docs scores
    scores = {}

    # ranking according to TF-IDF
    for file in files:
        tf_idf = []
        # get all words in file
        lstwords = files.get(file)
        # iterate over words in query to count amount of times its in the doc
        for word in query:
            tf = 0
            for wrd in lstwords:
                if wrd == word:
                    # get all matches
                    tf += 1
            # calculate tf_idf using TF and idf
            tf_idf.append(tf * idfs[word])
        # add file with sum of all query words' tf_idf score
        scores[file] = sum(tf_idf)

    # sort with the highest value first
    sorted_scores = sorted(scores.items(), key=lambda item: item[1], reverse=True)

    # add as many of the top files to the list as required by n
    for i in range(n):
        filenames.append(sorted_scores[i][0])

    return filenames


def top_sentences(query, sentences, idfs, n):
    """
    Given a `query` (a set of words), `sentences` (a dictionary mapping
    sentences to a list of their words), and `idfs` (a dictionary mapping words
    to their IDF values), return a list of the `n` top sentences that match
    the query, ranked according to idf. If there are ties, preference should
    be given to sentences that have a higher query term density.
    """

    # initialise final list
    final = []

    # dict to keep track of sentence scores
    scores = {}

    for sentence in sentences:
        # matching word measure
        mwm = []
        for word in sentences[sentence]:
            for wrd in query:
                if word == wrd:
                    mwm.append(idfs[word])

        scores[sentence] = sum(mwm)

    # sort with the highest value first
    sorted_scores = sorted(scores.items(), key=lambda item: item[1], reverse=True)

    # add as many of the top files to the list as required by n
    for i in range(n):
        final.append(sorted_scores[i][0])

    return final


if __name__ == "__main__":
    main()
